from typing import Union

import networkx as nx
from concurrent.futures import ThreadPoolExecutor, as_completed
import math

import numpy as np

from semistaticsim.rendering.floor import ReachabilityGraph
from semistaticsim.rendering.utils.dict2xyztup import dict2xyztuple


def cleanup_request(requested_node):
    if isinstance(requested_node, dict) and all([x in requested_node for x in ['x', 'y', 'z']]):
        requested_node = dict2xyztuple(requested_node)
    if isinstance(requested_node, tuple):
        requested_node = list(requested_node)
    return requested_node

def astar(requested_start_node, requested_end_node, requested_end_size, filtered_rg: ReachabilityGraph):
    requested_start_node = cleanup_request(requested_start_node)
    requested_end_node = cleanup_request(requested_end_node)

    # filtered_rg = full_reachability_graph.filter_(controller.last_event.metadata["objects"], mindist_to_obj=0.5)

    # filtered_rg.parent_floor.plot_self(reachability_graph=filtered_rg)
    # filtered_rg.parent_floor.plot_self(reachability_graph=filtered_rg)
    # filtered_rg.parent_floor.plot_self(samples=jnp.array(path))

    start_nodes, start_distances = filtered_rg.shunt_to_bbox(requested_start_node, None)
    end_nodes, end_distances = filtered_rg.shunt_to_bbox(requested_end_node, requested_end_size)

    for n in start_nodes + end_nodes:
        assert n in filtered_rg.full_graph

    start_node = start_nodes[start_distances.argmin()]

    # 5. Define the worker function for the thread pool
    def get_path_safe(target_node):
        try:
            # We use weight='weight' if your edges have weights, otherwise it defaults to 1
            # You can also pass a heuristic function here if you have one
            path = nx.astar_path(filtered_rg.full_graph, start_node, target_node)

            # Return slice [1:] as requested (excluding the start node)
            return path[1:]
        except nx.NetworkXNoPath:
            return None

    valid_paths = []

    # 6. Execute pathfinding in parallel
    # Using ThreadPoolExecutor is usually best for graph traversals as they release the GIL
    # and we avoid the overhead of pickling the whole graph for ProcessPoolExecutor.
    with ThreadPoolExecutor() as executor:
        # Submit all tasks
        future_to_node = {executor.submit(get_path_safe, node): node for node in end_nodes}

        for future in as_completed(future_to_node):
            path = future.result()
            if path is not None:
                valid_paths.append(path)

    if not valid_paths:
        # Handle the case where no path could be found to ANY end node
        raise ValueError(f"No path found from {start_node} to any of the {len(end_nodes)} candidate end nodes.")

    # 7. Select the best path
    # Currently selecting the path with the fewest hops/lowest weight.
    # If you want to account for 'end_distances', you would zip those values in earlier.
    best_path = min(valid_paths, key=lambda x: len(x))

    return best_path

def geodesic_distance_on_path_nodes(path_nodes, grid_size: Union[ReachabilityGraph, float, int]):
    if isinstance(grid_size, int) or isinstance(grid_size, float):
        pass
    else:
        grid_size = grid_size.grid_size
    if not isinstance(path_nodes, list):
        path_nodes = list(path_nodes)

    if len(path_nodes) == 0:
        return 0

    if len(path_nodes) == 1:
        return grid_size

    euclidean_distance = sum(
        np.linalg.norm(np.array(a) - np.array(b))
        for a, b in zip(path_nodes[:-1], path_nodes[1:])
    ) + grid_size
    return euclidean_distance

def geodesic_distance(requested_start_node, requested_end_node, requested_end_size, filtered_rg: ReachabilityGraph):
    return geodesic_distance_on_path_nodes(astar(requested_start_node, requested_end_node, requested_end_size, filtered_rg), filtered_rg.grid_size)


import shutil
import numpy as np
import os
import jax.numpy as jnp

from pathlib import Path
from tqdm import tqdm
import json

from semistaticsim.datawrangling.sssd import load_sssd, GeneratedSemiStaticData

def purge_dir(res_path, prefix):
    for item in os.listdir(res_path):
        item_path = os.path.join(res_path, item)
        if not os.path.isdir(item_path) or not item.startswith(prefix):
            if os.path.isdir(item_path):
                shutil.rmtree(item_path)
            else:
                os.remove(item_path)


def move_indexed_files(src_dir: Path, dst_dir: Path, indices, ext):
    dst_dir.mkdir(parents=True, exist_ok=True)
    for trg_idx, src_idx in enumerate(indices):
        shutil.move(src_dir / f"{src_idx:05d}.{ext}", dst_dir / f"{trg_idx:05d}.{ext}")


def save_run(
    mask: np.ndarray,
    sssd_data: GeneratedSemiStaticData,
    res_path: str,
    tmpdir: str,
    batch_size: int,
    gt_sss_data: GeneratedSemiStaticData,
):
    # Delete existing results
    if os.path.exists(res_path):
        shutil.rmtree(res_path)
    os.makedirs(res_path, exist_ok=True)
    # Index df samples
    indices = np.nonzero(mask)[0]
    subset = sssd_data.take_by_global_timestep(stop_or_start=None, indices=indices)

    # --- Move Images ---
    images_src = tmpdir / "images"
    images_dst = res_path / "images"
    for subdir in images_src.iterdir():
        if subdir.is_dir():
            # detect extension from first file
            sample = next(subdir.iterdir())
            ext = sample.suffix.lstrip(".")

            move_indexed_files(src_dir=subdir, dst_dir=images_dst / subdir.name, indices=indices, ext=ext)

    # --- Move poses ---
    move_indexed_files(src_dir=tmpdir / "poses", dst_dir=res_path / "poses", indices=indices, ext="json")

    # Save parquet
    subset.dump_to_parquet(res_path, dump_leftover=True, verbose=False, batch_size=batch_size)

    # Move keyframes txt and fix indices
    keyframes_src = tmpdir / "keyframes.txt"
    with open(keyframes_src, "r") as f:
        keyframe_lines = f.readlines()
    keyframe_indices = [int(line.strip()) for line in keyframe_lines]
    # Map from global indices to run indices
    index_map = {global_idx: run_idx for run_idx, global_idx in enumerate(indices)}
    keyframe_run_indices = [index_map[idx] for idx in keyframe_indices if idx in index_map]
    keyframes_dst = res_path / "keyframes.txt"
    with open(keyframes_dst, "w") as f:
        for idx in keyframe_run_indices:
            f.write(f"{idx:05d}\n")

    # Copy id_to_color mapping if exists
    id_to_color_src = tmpdir / "id_to_color.json"
    if id_to_color_src.exists():
        shutil.copy(id_to_color_src, res_path / "id_to_color.json")

    def accuracy(subset):
        preds: jnp.array = subset.sum_up_minors_into_major()._assignment
        targets: jnp.array = gt_sss_data._assignment
        # Only get actual predictions
        valid_mask = preds >= 0
        matches = (preds == targets) & valid_mask
        # Global accuracy
        total_correct = matches.sum()
        total_valid = valid_mask.sum()
        global_acc = jnp.where(total_valid > 0, total_correct / total_valid, 0.0)
        # Object accuracy
        obj_correct = matches.sum(axis=-1)
        obj_valid = valid_mask.sum(axis=-1)
        object_acc = jnp.where(obj_valid > 0, obj_correct / obj_valid, 0.0)
        return global_acc.item(), object_acc.squeeze().tolist()

    global_acc, object_acc = accuracy(subset)
    # Save accuracy as json
    with open(os.path.join(res_path, "accuracy.json"), "w") as f:
        json.dump({"global_accuracy": global_acc, "object_accuracy": object_acc}, f)


def format_results(res_path: str, tmpdir: str, batch_size: int, gt_sss_data: GeneratedSemiStaticData) -> dict:
    sssd_data = load_sssd(tmpdir)
    # Get timestamps to split data folder
    timestamps = np.concatenate([data._timestamp for data in sssd_data.get_generator_of_selves()])
    floored_timestamps = np.floor(timestamps)
    unique_timestamps = np.unique(floored_timestamps)
    # Split data into multiple folders based on timestamps
    for t in tqdm(unique_timestamps, desc="Formatting results"):
        mask = floored_timestamps == t
        if mask.sum() > 1:
            print(f"Processing run_id: {int(t.item())} with {mask.sum()} samples")
            gt_data_for_this_major = gt_sss_data.take_by_global_timestep(int(t), int(t) + 1)
            save_run(
                mask,
                sssd_data,
                res_path=Path(res_path) / f"run_{int(t.item())}",
                tmpdir=Path(tmpdir),
                batch_size=batch_size,
                gt_sss_data=gt_data_for_this_major,
            )



from abc import ABC
from typing import List, Dict, Union, Self, Any

import flax
import numpy as np
from flax import struct

from semistaticsim.datawrangling.sssd import GeneratedSemiStaticData
from semistaticsim.keyboardcontrol.main_skillsim import ROBOTS
from semistaticsim.rendering.simulation.skill_simulator import Simulator

class Agent(ABC):
    def __init__(self, sim: Simulator):
        self.sim = sim

    @property
    def current_time(self):
        return self.sim.sss_data.pickupable_selves_at_current_time._timestamp[0]

    def _resolve_query_into_pickupable(self, query: str, current_time: float) -> List[str]:
        """

        Args:
            query: An open-vocab query specifying a target pickupable.

        Returns: The top N most likely pickupables, sorted by likelihood (most likely is at index 0)

        """
        raise NotImplementedError()

    def resolve_query_into_pickupable(self, query: str, current_time: float=None) -> List[str]:
        if current_time is None:
            current_time = self.current_time
        return self._resolve_query_into_pickupable(query, current_time)

    def _resolve_pickupable_into_weighted_receptacles(self, pickupable: str, current_time: float = None) -> Dict[str, float]:
        """

        Args:
            pickupable: A requested pickupable.
            current_time:

        Returns: A probability weight for each receptacle for the current location of the pickupable.

        """
        raise NotImplementedError()

    def resolve_pickupable_into_weighted_receptacles(self, pickupable: str, current_time: float=None) -> Dict[str, float]:
        if current_time is None:
            current_time = self.current_time

        return self._resolve_pickupable_into_weighted_receptacles(pickupable, current_time)

    def update(self, observation: Dict[str, Any]) -> Union[None, Self]:
        """

        Args:
            observation: An observation of the current state.

        Returns: A new Agent instance that has been updated given the observation.

        """
        raise NotImplementedError()

    def found_pickupable(self, pickupable_name: str, observation: Dict[str, Any]) -> bool:
        """

        Args:
            observation: An observation of the current state.

        Returns: True if the current state contains the pickupable, False otherwise.

        """
        raise NotImplementedError()

    def goto_query(self, query: str, current_time: float=None, custom_traversal=None) -> List[str]:
        pickupables = self.resolve_query_into_pickupable(query, current_time)

        pickupable = pickupables[0]
        weighted_receptacles = self.resolve_pickupable_into_weighted_receptacles(pickupable, current_time)
        sorted_keys = sorted(weighted_receptacles, key=weighted_receptacles.get, reverse=True)

        if custom_traversal is not None:
            return custom_traversal(sorted_keys)

        for k in sorted_keys:
            self.sim.GoToObject(ROBOTS[0], k, max_path_length=2)
            obs = self.sim.render()
            self.sim.privileged_apn = None

            if self.found_pickupable(pickupable, obs):
                return obs
            self.update(obs)
